# Copyright 2010 Neil Martinsen-Burrell

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 

from bzrlib.plugins.colo.tests.script import ScriptTestCase
from bzrlib.plugins.colo.colocated import ColocatedWorkspace


class TestColoSyncFrom(ScriptTestCase):

    def test_sync(self):
        self.run_script("""
$ bzr colo-init remote
$ bzr colo-init local
$ cd local
$ bzr colo-sync-from ../remote
2>Updating branch trunk from colo:../remote:trunk.
No revisions to pull.
""")

    def test_sync_nontrivial(self):
        self.run_script("""
$ bzr colo-init remote
$ bzr colo-init local
$ cd remote
$ echo content > file
$ bzr add
$ bzr ci -m 'a revision'
$ cd ../local
$ bzr colo-sync-from ../remote
2>Updating branch trunk from colo:../remote:trunk.
2>All changes applied successfully.
Now on revision 1.
""")

    def test_sync_missing_branch(self):
        self.run_script("""
$ bzr colo-init remote
$ bzr colo-init local
$ cd local
$ bzr colo-branch other
$ bzr colo-sync-from ../remote
$ bzr colo-branches
* other
  trunk
""")

    def test_sync_creates_branch(self):
        self.run_script("""
$ bzr colo-init remote
$ bzr colo-init local
$ cd remote
$ bzr colo-branch other
$ cd ../local
$ bzr colo-sync-from ../remote
$ bzr colo-branches
  other
* trunk
""")


    def test_sync_diverged(self):
        self.run_script("""
$ bzr colo-init remote
$ bzr colo-init local
$ cd remote
$ echo content > file
$ bzr add
$ bzr ci -m 'a revision in remote branch'
$ cd ../local
$ echo stuff > file1
$ bzr add
$ bzr ci -m 'a different revision in local branch'
""")
        self.run_bzr_error(['These branches have diverged'],
                           ['colo-sync-from', '../remote'])

    def test_sync_multiple_diverged(self):
        self.run_script("""
$ bzr colo-init local
$ bzr colo-init remote
$ cd remote
$ echo content > file
$ bzr add
$ bzr ci -m 'trunk revision'
$ bzr colo-branch other
$ cd ../local
$ echo stuff > file
$ bzr add
$ bzr ci -m 'a different trunk revision'
$ bzr colo-sync-from ../remote
2>Creating new branch other from colo:../remote:other.
2>Branched 1 revision(s).
2>Updating branch trunk from colo:../remote:trunk.
2>bzr: ERROR: These branches have diverged...
""")

    def test_sync_overwrite(self):
        self.run_script("""
$ bzr colo-init remote
$ bzr colo-init local
$ cd remote
$ echo content > file
$ bzr add
$ bzr ci -m 'a revision in remote branch'
$ cd ../local
$ echo stuff > file1
$ bzr add
$ bzr ci -m 'a different revision in local branch'
$ bzr colo-sync-from --overwrite ../remote
2>Updating branch trunk from colo:../remote:trunk.
2>All changes applied successfully.
Now on revision 1.
""")

    def test_sync_from_empty(self):
        self.run_script("""
$ mkdir empty
$ bzr colo-init local
$ cd local
$ bzr colo-sync-from ../empty
2>bzr: ERROR: No colocated workspace in ../empty
""")

    def test_sync_from_standalone(self):
        self.run_script("""
$ bzr colo-init local
$ bzr init standalone
$ cd local
$ bzr colo-sync-to ../standalone
2>bzr: ERROR: No colocated workspace in ../standalone
""")


class TestColoSyncTo(ScriptTestCase):

    def test_sync(self):
        self.run_script("""
$ bzr colo-init remote
$ bzr colo-init local
$ cd local
$ bzr colo-sync-to ../remote
2>Updating colo:../remote:trunk from trunk.
2>No new revisions to push.
""")

    def test_sync_nontrivial(self):
        self.run_script("""
$ bzr colo-init remote
$ bzr colo-init local
$ cd local
$ echo content > file
$ bzr add
$ bzr ci -m 'first revision'
$ bzr colo-sync-to ../remote
2>Updating colo:../remote:trunk from trunk.
2>Pushed up to revision 1.
$ cd ../remote
$ bzr revno
1
""")

    def test_sync_to_empty(self):
        self.run_script("""
$ mkdir branch
$ bzr colo-init workspace
$ cd workspace
$ bzr colo-sync-to ../branch
2>bzr: ERROR: No colocated workspace in .../branch
""")

    def test_sync_to_standalone(self):
        self.run_script("""
$ bzr colo-init local
$ bzr init standalone
$ cd local
$ bzr colo-sync-to ../standalone
2>bzr: ERROR: No colocated workspace in ../standalone
""")

    def test_sync_to_no_trees(self):
        self.run_script("""
$ bzr colo-init --no-tree remote
$ bzr colo-init local
$ cd local
$ bzr colo-sync-to ../remote
2>Updating colo:../remote:trunk from trunk.
2>No new revisions to push.
$ cd ../remote
$ bzr colo-branches
  trunk
""")

    def test_sync_to_creates_branch(self):
        self.run_script("""
$ bzr colo-init remote
$ bzr colo-init local
$ cd local
$ bzr colo-branch other
$ bzr colo-sync-to ../remote
2>Updating colo:../remote:other from other.
2>Created new branch.
2>Updating colo:../remote:trunk from trunk.
2>No new revisions to push.
$ cd ../remote
$ bzr colo-branches
  other
* trunk
""")

    def test_sync_to_diverged(self):
        self.run_script("""
$ bzr colo-init remote
$ bzr colo-init local
$ cd remote
$ echo content > file
$ bzr add
$ bzr ci -m 'a revision'
$ cd ../local
$ echo stuff > file
$ bzr add
$ bzr ci -m 'a separate revision'
$ bzr colo-sync-to ../remote
2>Updating colo:../remote:trunk from trunk.
2>bzr: ERROR: These branches have diverged. ...
""")


class TestColoSyncLocations(ScriptTestCase):

    def test_set_locations(self):
        self.run_script('$ bzr colo-init')
        w = ColocatedWorkspace()
        c = w.get_config()
        self.assertEqual(c.get_sync_from_location(), None)
        self.assertEqual(c.get_sync_to_location(), None)
        c.set_sync_from_location('location')
        c.set_sync_to_location('other_location')
        self.assertEqual(c.get_sync_from_location(), 'location')
        self.assertEqual(c.get_sync_to_location(), 'other_location')

    def test_set_location_none(self):
        self.run_script('$ bzr colo-init')
        w = ColocatedWorkspace()
        c = w.get_config()
        c.set_sync_from_location(None)
        self.assertEqual(c.get_sync_from_location(), None)
        c.set_sync_to_location(None)
        self.assertEqual(c.get_sync_to_location(), None)

    def test_sync_from_default_location(self):
        self.run_script("""
$ bzr colo-init remote
$ bzr colo-init local
$ cd local
$ bzr colo-sync-from ../remote
$ bzr colo-sync-from
2>Updating branch trunk from .../remote/:trunk.
No revisions to pull.
""")

    def test_sync_from_remembered_location(self):
        self.run_script("""
$ bzr colo-init remote1
$ bzr colo-init remote2
$ bzr colo-init local
$ cd local
$ bzr colo-sync-from ../remote1
2>Updating branch trunk from ...:trunk.
No revisions to pull.
$ bzr colo-sync-from --remember ../remote2
$ bzr colo-sync-from
2>Updating branch trunk from ...remote2/:trunk.
""")

    def test_sync_from_no_default(self):
        self.run_script("""
$ bzr colo-init
$ bzr colo-sync-from
2>bzr: ERROR: No synchronization location known or specified.
""")

    def test_sync_to_default_location(self):
        self.run_script("""
$ bzr colo-init remote
$ bzr colo-init local
$ cd local
$ bzr colo-sync-to ../remote
$ bzr colo-sync-to
2>Updating colo:.../remote/:trunk from trunk.
2>No new revisions to push.
""")

    def test_sync_to_remembered_location(self):
        self.run_script("""
$ bzr colo-init remote1
$ bzr colo-init remote2
$ bzr colo-init local
$ cd local
$ bzr colo-sync-to ../remote1
2>Updating colo:.../remote1:trunk from trunk.
2>No new revisions to push.
$ bzr colo-sync-to --remember ../remote2
$ bzr colo-sync-to
2>Updating colo:.../remote2/:trunk from trunk.
2>No new revisions to push.
""")

    def test_sync_to_no_default(self):
        self.run_script("""
$ bzr colo-init
$ bzr colo-sync-to
2>bzr: ERROR: No synchronization location known or specified.
""")



# Copyright 2009 Neil Martinsen-Burrell
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 

import sys
import bzrlib
from bzrlib.lazy_import import lazy_import
lazy_import(globals(), '''
from bzrlib import (api,
                    branch as _mod_branch,
                    builtins,
                    bzrdir,
                    commands,
                    directory_service,
                    errors, 
                    option,
                    osutils,
                    reconfigure,
                    switch,
                    trace, 
                    transport,
                    urlutils, 
                    workingtree,
                   )
from bzrlib.repofmt import groupcompress_repo, pack_repo
''')

from colocated import (COLOCATED_LOCATION,
                       ColocatedWorkspace,
                       NoCurrentBranch,
                      )

COLOCATED_STAGING = u'.bzrbranches'

def _ensure_base_of_location(to_transport, create_parent_dirs):
    try:
        to_transport.ensure_base()
    except errors.NoSuchFile:
        if not create_parent_dirs:
            raise
        to_transport.create_prefix()


def _wrap_command_object(cmd_obj, outf):
    # We explicitly reuse the current command's output stream rather than
    # setup a fresh one here
    cmd_obj.outf = outf
    if (2,0) < api.get_current_api_version(bzrlib) < (2,2):
        trace.mutter('Using 2.1 compatibility')
        cmd_obj.run = cmd_obj.run_direct
    return cmd_obj
    

class HideStagingWrapper(object):

    """Hide the staging directory in output."""

    def __init__(self, stream):
        self.stream = stream

    def __getattr__(self, name):
        return getattr(self.stream, name)

    def _hide(self, input_string):
        return input_string.replace(COLOCATED_STAGING,
                                    COLOCATED_LOCATION
                                   )

    def write(self, input_string, *args, **kwargs):
        self.stream.write(self._hide(input_string),
                          *args, **kwargs)

    def writelines(self, line_list, *args, **kwargs):
        self.stream.writelines(self._hide(line) for line in line_list)


def init_workspace(location, format=None, create_parent_dirs=False,
                   outf=None, branch_name='trunk', create_tree=True):
    """Create a colocated workspace at the specified location.
    
    :param location: A path or URL for where to make the workspace
    :param format: A Bazaar format to use for the branches and trees
    :param create_parent_dirs: Whether or not to create the directories
      containing the workspace
    """
    if outf is None:
        outf = sys.stdout
    if format is None:
        format = bzrdir.format_registry.get('default')()
    
    # if the repository is going to live inside of the checkout's .bzr,
    # we need to create the checkout's .bzr first since if that can't happen
    # then we don't want to create the repository under that.  But, we can't
    # create a checkout since we don't have a branch to check out.  Instead
    # we put the repository in a staging location
    top_transport = transport.get_transport(location)
    _ensure_base_of_location(top_transport, create_parent_dirs)
    topdir = format.initialize_on_transport(top_transport)

    repo_transport = top_transport.clone(COLOCATED_LOCATION)
    repo_transport.ensure_base()
    repodir = format.initialize_on_transport(repo_transport)
    repo = repodir.create_repository(shared=True)
    repo.set_make_working_trees(False)

    branch_transport = repo_transport.clone(branch_name)
    branch_transport.create_prefix()
    branch_transport.ensure_base()
    branchdir = format.initialize_on_transport(branch_transport)
    branch = branchdir.create_branch()

    if create_tree:
        from_branch = _mod_branch.BranchReferenceFormat().initialize(
            topdir, target_branch=branch)
        topdir.create_workingtree(from_branch=from_branch)
        


class cmd_colo_init(commands.Command):

    """Create a working tree with colocated branches.

    A colocated workspace is a working tree that also contains all of its
    branches within the same directory.  The branches are stored in the
    .bzr/branches subdirectory.  They may be easily referred to using the 
    colo: specifier.

    By default, the initial branch is called "trunk".  A different name can 
    be specified with the --trunk-name option.  For creating colocated
    workspaces remotely, a --no-tree options is available for situations
    where it isn't possible or desired to create a working tree.
    """

    takes_options = [
        option.Option('create-prefix',
           help='Create the path leading up to the location '
                'if it does not already exist.'),
        option.RegistryOption('format',
                help='Specify a format for the repository, etc. '
                'See "help formats".',
                lazy_registry=('bzrlib.bzrdir', 'format_registry'),
                converter=lambda name: bzrdir.format_registry.make_bzrdir(name),
                value_switches=True,
                title="Storage Format",
                ),
        option.Option('trunk-name',
                      help='Name for the trunk branch',
                      type=unicode),
        option.Option('no-tree',
                      help='Create a workspace without a working tree'),
        ]
    takes_args = ['location?']
    
    def run(self, location=u'.', format=None, create_prefix=False,
            trunk_name='trunk', no_tree=False):
        if format is None:
            format = bzrdir.format_registry.make_bzrdir('default')
        init_workspace(location, format=format, 
                       create_parent_dirs=create_prefix,
                       outf=self.outf, branch_name=trunk_name,
                       create_tree=(not no_tree)
                      )


class cmd_colo_branch(commands.Command):

    """Make a new branch from the currently active branch.

    In a colocated workspace, there is always one branch active at a time,
    so to create a new branch, we need to only specify the name of the new
    branch.  If the ``-r`` option is given, then the branch is made from
    the specified revision of the current branch.  By default, the branch
    is made from the current branch, but a different branch name to branch
    from can be specified with the ``--from-branch`` option.
    """

    takes_options = ['revision',
        option.Option('no-switch',
               help="Don't switch the workspace to the new branch"),
        option.Option('from-branch',
               type=unicode,
               help='name of the branch to branch from'),
                    ]
    takes_args = ['new_branch']

    def run(self, new_branch, revision=None, no_switch=False,
            from_branch=None):
        if not new_branch.startswith('colo:'):
            new_branch = 'colo:' + new_branch
        new_location = directory_service.directories.dereference(new_branch)
        transport.get_transport(new_location).create_prefix()
        workspace = ColocatedWorkspace()
        if not workspace.has_tree:
            no_switch = True
        if from_branch is not None:
            from_branch = workspace.branch_with_name(from_branch)
        else:
            from_branch = workspace.current_branch()
        branch_cmd = _wrap_command_object(builtins.cmd_branch(), self.outf)
        branch_cmd.run(from_branch.base, revision=revision,
                       to_location=new_location, no_tree=True,
                       switch=(not no_switch), use_existing_dir=True)


class cmd_colo_branches(commands.Command):

    """List all of the branches in a colocated workspace.
    
    If the current directory is a working tree pointing to one of the branches,
    that branch is marked with a "*".
    """

    takes_args = ['location?']
    
    def run(self, location=u'.'):
        workspace = ColocatedWorkspace(location)
        current_name = workspace.current_branch_name()
        for name in workspace.branch_names():
            leader = '  '
            if name == current_name:
                leader = '* '
            self.outf.write(leader + name + '\n')


class cmd_colo_fetch(commands.Command):
    
    """Fetch an external project into a new colocated workspace.
    
    This creates a colocated workspace at the specified location with a single
    branch named "origin/trunk" that has the contents of the external project.
    If you would like to specify a different name for the new branch, use the
    ``--trunk-name`` option.  If you don't specify a TO_LOCATION, then one
    will be derived from the final component of the FROM_LOCATION (similar
    to ``bzr branch``.
    
    """

    takes_args = ['from_location', 'to_location?']
    takes_options = [
        option.Option('create-prefix',
           help='Create the path leading up to the location '
                'if it does not already exist.'),
        option.RegistryOption('format',
                help='Specify a format for the repository, etc. '
                'See "help formats".',
                lazy_registry=('bzrlib.bzrdir', 'format_registry'),
                converter=lambda name: bzrdir.format_registry.make_bzrdir(name),
                value_switches=True,
                title="Storage Format",
                ),
        option.Option('trunk-name',
            help='Name of the branch to create; default is "origin/trunk"',
            type=unicode),
        ]

    def run(self, from_location=None, to_location=None,
            create_prefix=False, format=None, trunk_name='origin/trunk'):
        if to_location is None:
            to_location = urlutils.derive_to_location(from_location)
        init_workspace(to_location, format=format, 
                       create_parent_dirs=create_prefix,
                       outf=self.outf, branch_name=trunk_name,
                      )
        pull_cmd = _wrap_command_object(builtins.cmd_pull(), self.outf)
        pull_cmd.run(location=from_location, directory=to_location,
                     remember=True)


class cmd_colo_checkout(commands.Command):

    """Create a checkout of this branch in a new directory.
    
    By default this command makes a new checkout of the current branch in
    the specified directory.  Using the ``--create`` option, a new branch
    named after the last component of the new directory is created, and
    the new checkout is switched to that new branch.
    """

    aliases = ['colo-co']
    takes_args = ['to_directory']
    takes_options = [
        option.Option('create', short_name='b',
               help="Also create a new colocated branch by this name"),
        option.Option('from-branch',
               type=unicode,
               help='name of the branch to branch from'),
        ]

    def run(self, to_directory, create=False, from_branch=None):
        workspace = ColocatedWorkspace()
        if from_branch is not None:
            from_branch = workspace.branch_with_name(from_branch)
        else:
            from_branch = workspace.current_branch()
        current_branch_url = from_branch.base
        if create:
            new_branch_name = osutils.basename(to_directory)
            _wrap_command_object(cmd_colo_branch(), self.outf).run(
                new_branch_name, no_switch=True)
            branch_location = urlutils.join(current_branch_url, '..',
                                            new_branch_name)
        else:
            branch_location = current_branch_url
        # create a new checkout
        checkout_cmd = _wrap_command_object(builtins.cmd_checkout(),
                                            self.outf)
        checkout_cmd.run(branch_location, to_location=to_directory,
                         lightweight=True)


class cmd_colo_prune(commands.Command):
    
    """Remove a branch from the colocated workspace.
    
    This commands removes the specified branches from the colocated workspace
    by removing them from the filesystem.  By default, it does not remove the
    revisions from the shared repository.  If it is desired to remove the
    revisions that are only present in the removed branches, then the
    ``--clean`` option may be given.
    """

    aliases = ['colo-delete']
    takes_options = [option.Option('clean',
                            help='Clean the repository after pruning.'),
                    ]
    takes_args = ['branch*']
    _see_also = ['colo-clean']

    def run(self, branch_list=None, clean=False):
        if branch_list is None:
            return
        if len(branch_list) < 1:
            raise errors.BzrOptionError('Must specify a branch to prune.')
        workspace = ColocatedWorkspace()
        all_branches = list(workspace.branch_names())
        to_transport = workspace.repo_transport()
        current_name = workspace.current_branch_name()
        for branch_name in branch_list:
            if branch_name.startswith('colo:'):
                branch_name = branch_name[5:]
            if branch_name == current_name:
                trace.note('Not removing current branch %s.' % branch_name)
            elif branch_name not in all_branches:
                trace.note('Not removing %s because it is not a branch.' %
                           branch_name)
            else:
                to_transport.delete_tree(branch_name)
        if clean:
            _clean(self.outf)


class cmd_colo_pull(commands.Command):

    """Update all remote branches.

    A remote branch is one whose name starts with "origin/".  This command
    updates all such branches by doing a pull from their default pull
    location.  To set this location, do ``bzr switch colo:origin/branch_name``
    and ``bzr pull --remember URL``.
    """

    takes_options = [option.Option('directory', short_name='d',
                                   help='colocated workspace to use',
                                   type=unicode),
                    ]
    
    def run(self, directory=u'.'):
        conflicts = 0
        workspace = ColocatedWorkspace(directory)
        current = workspace.current_branch_name()
        for name in (_ for _ in workspace.branch_names()
                       if _.startswith('origin/')):
            br = workspace.branch_with_name(name)
            parent = br.get_parent()
            if parent is None:
                continue
            trace.note('Updating %s from %s' % 
                       (name, urlutils.unescape_for_display(parent,
                                                            self.outf.encoding
                                                           )))
            parent_branch = _mod_branch.Branch.open(parent)
            if parent_branch is not br: # can a branch be its own parent?
                                        # cribbed from bzrlib/builtins.py
                parent_branch.lock_read()
                self.add_cleanup(parent_branch.unlock)
            
            br.lock_write()
            self.add_cleanup(br.unlock)
            result = br.pull(parent_branch)
            self.cleanup_now()

            result.report(self.outf)
            if name == current:
                # update the working tree
                tree = workingtree.WorkingTree.open(directory)
                conflicts = tree.update()
        if conflicts > 0:
            return 1
        else:
            return 0


class cmd_colo_mv(commands.Command):

    """Rename a colocated branch."""

    takes_args = ['oldname', 'newname']
    aliases = ['colo-move', 'colo-rename']

    def run(self, oldname, newname):
        workspace = ColocatedWorkspace()
        current = workspace.current_branch_name()
        repo_transport = workspace.repo_transport()
        to_transport = repo_transport.clone(newname)
        to_transport.create_prefix()
        repo_transport.move(oldname, newname)
        if current == oldname:  #switch checkout to new name
            this_bzrdir, _ = bzrdir.BzrDir.open_containing(u'.')
            switch.switch(this_bzrdir, 
                          workspace.branch_with_name(newname),
                          force=True)


class NotPackRepository(errors.BzrError):

    _fmt = "Can't clean the repository at %(repo)s"

    def __init__(self, repo):
        self.repo = repo


def _clean(outf, directory=u'.'):
    """Clean a colocated workspace's repository"""
    workspace = ColocatedWorkspace(directory)
    repo_path = urlutils.local_path_from_url(workspace.repo_location)
    outf.write("Cleaning repository in %s.\n" % repo_path)
    repo = workspace.repository()
    try:
        pack_collection = repo._pack_collection
    except AttributeError:
        raise NotPackRepository(repo_path)
    repo.lock_read()
    try:
        pack_collection.ensure_loaded()
    finally:
        repo.unlock()

    # Keep all of the revision ids in the ancestry of the current heads
    ancestry = set()
    for b in workspace.branches():
        b.lock_read()
        try:
            ancestry.update(set(rev[0] for rev in 
                b.iter_merge_sorted_revisions(b.last_revision())))
        finally:
            b.unlock()

    old_packs = pack_collection.all_packs()
    if isinstance(repo._format, (groupcompress_repo.RepositoryFormatCHK1,
                                 groupcompress_repo.RepositoryFormatCHK2,
                                 groupcompress_repo.RepositoryFormat2a,)):
        packer_class = groupcompress_repo.GCCHKPacker
    elif isinstance(repo._format, 
                    (pack_repo.RepositoryFormatKnitPack1,
                     pack_repo.RepositoryFormatKnitPack3,
                     pack_repo.RepositoryFormatKnitPack4,
                     pack_repo.RepositoryFormatKnitPack5RichRoot,
                     pack_repo.RepositoryFormatKnitPack5RichRootBroken,
                     pack_repo.RepositoryFormatKnitPack6,
                     pack_repo.RepositoryFormatKnitPack6RichRoot,
                     pack_repo.RepositoryFormatPackDevelopment2Subtree,)):
        packer_class = pack_repo.Packer
    else:
        trace.show_error("Can't clean repository of type %s",
                         repo.get_format_string()) 
    packer = packer_class(pack_collection,
                          old_packs,
                          suffix='.clean', 
                          revision_ids=ancestry)
    repo.lock_write()
    try:
        result = packer.pack()
    finally:
        repo.unlock()

    if result is None:
        trace.note('No cleaning necessary.')
        return

    for pack in old_packs:
        pack_collection._remove_pack_from_memory(pack)
    pack_collection._save_pack_names(clear_obsolete_packs=True)
    pack_collection._obsolete_packs(old_packs)


class cmd_colo_clean(commands.Command):
    
    """Clean unused revisions from the repository.

    An unused revision is one that is not in the ancestry of any of the 
    active branches in the workspace.  This command can be used after the
    ``colo-prune`` command to remove revisions that were only present in
    the branch that was pruned.
    """

    def run(self):
        _clean(self.outf)


class cmd_colo_ify(commands.Command):

    """Convert an existing branch into a colocated workspace.

    This converts an existing branch into a colocated workspace.  Other
    branches can be added to the workspace by giving them as arguments.  The
    names of the other branches will be determined from their nicknames.  To
    use a different name for the current branch in the colocated workspace,
    use the ``--trunk-name`` option.
    """

    _see_also = ['nick']
    takes_args = ['other_location*']
    takes_options = [option.Option('trunk-name',
                                   help='The name to use for this branch in '
                                        'the colocated workspace',
                                   type=unicode),
                    ]

    def run(self, other_location_list=[], trunk_name='trunk'):
        the_bzrdir, relpath = bzrdir.BzrDir.open_containing(u'.')
        the_bzrdir.root_transport.mkdir(COLOCATED_LOCATION)
        repo_location = urlutils.join(the_bzrdir.root_transport.base,
                          COLOCATED_LOCATION)
        new_transport = transport.get_transport(repo_location,
            possible_transports=[the_bzrdir.root_transport])
        new_bzrdir = the_bzrdir.cloning_metadir().initialize_on_transport(
                         new_transport)
        new_repo = new_bzrdir.create_repository(shared=True)
        new_repo.set_make_working_trees(False)

        the_branch = _mod_branch.Branch.open_containing(u'.')[0]
        branch_location = urlutils.join(repo_location, trunk_name)
        new_branch_bzrdir = the_bzrdir.sprout(branch_location,
                                       source_branch=the_branch,
                                       create_tree_if_local=False)
        new_branch = new_branch_bzrdir.open_branch()
        new_branch.set_parent(None)
        self._copy_locations(the_branch, new_branch)

        reconfiguration = reconfigure.Reconfigure.to_lightweight_checkout(
                              the_bzrdir, reference_location=branch_location)
        reconfiguration.apply()

        if other_location_list is None:
            return
        for location in other_location_list:
            source_branch = _mod_branch.Branch.open(location)
            to_location = urlutils.join(repo_location, source_branch.nick)
            new_bzrdir = source_branch.bzrdir.sprout(to_location,
                source_branch=source_branch,
                create_tree_if_local=False)
            self._copy_locations(source_branch, new_bzrdir.open_branch())

    def _copy_locations(self, source, dest):
        """Set all of the locations for one branch to those of another."""
        location_methods = [(source.get_parent, dest.set_parent),
                            (source.get_push_location,
                             dest.set_push_location),
                            (source.get_submit_branch,
                             dest.set_submit_branch),
                           ]
        for source_meth, dest_meth in location_methods:
            val = source_meth()
            if val is not None:
                dest_meth(val)


class cmd_colo_fixup(commands.Command):

    """Fix the checkout reference after moving a colocated workspace.

    Because Bazaar uses absolute pathnames in storing references to branches,
    when a colocated workspace is moved on disk, the link from the
    checkout to the current branch gets broken.  This command uses a simple
    heuristic to try to fix that reference.  If it doesn't work, then the
    command ``bzr switch --force .bzr/branches/<current branch name>`` will
    always work.
    """
    
    def run(self):
        # get the current location which is assumed to be wrong
        the_bzrdir = bzrdir.BzrDir.open_containing(u'.')[0]
        format = the_bzrdir.find_branch_format()
        if getattr(format, 'get_reference', None) is not None:
            location = format.get_reference(the_bzrdir)
        # find the root of the working tree
        root_location = the_bzrdir.root_transport.base
        orig, _, location_in_repo = location.partition(COLOCATED_LOCATION)
        try:
            format.set_reference(the_bzrdir,
                                 None,
                                 _mod_branch.Branch.open(
                                     urlutils.join(root_location,
                                         COLOCATED_LOCATION,
                                         location_in_repo.strip('/'))))
        except TypeError:
            # In 2.1, need to use one less argument
            format.set_reference(the_bzrdir,
                                 _mod_branch.Branch.open(
                                     urlutils.join(root_location,
                                         COLOCATED_LOCATION,
                                         location_in_repo.strip('/'))))


class NoQBzrError(errors.BzrCommandError):

    _fmt = '%(name)s requires the qbzr plugin version >= 0.18.1.'


class cmd_qbranches(commands.Command):

    """List colocated branches (requires qbzr)"""

    def run(self):
        raise NoQBzrError(name='qbranches')


class cmd_qprune(commands.Command):

    """Remove colocated branches (requires qbzr)"""

    aliases = ['qdelete']

    def run(self):
        raise NoQBzrError(name='qprune')


qbzr = None
try:
    import bzrlib.plugins.qbzr as qbzr
    import qcommands
except ImportError, e:
    trace.mutter("Can't import qbzr or some qbzr functions, "
        "not defining GUI commands.\n"
        "ImportError: %s" % str(e)
        )
else:
    cmd_qprune = qcommands.cmd_qprune
    cmd_qbranches = qcommands.cmd_qbranches


class NoSavedSyncLocation(errors.BzrCommandError):

    _fmt = ("No synchronization location known or specified.")

class cmd_colo_sync_from(commands.Command):

    """Synchronize this colocated workspace from the state of another one.
    
    LOCATION should be a path or URL to a colocated workspace such that the
    branches it holds are accessible as colo:LOCATION:branch_name.  For every
    branch in that remote workspace, if there is a local branch with the same 
    name, the remote branch is pulled into the corresponding local branch.  
    If there is not a local branch with a name, then a new branch is created
    in the local colocated workspace.

    The first time colo-sync-from is used, it records a default location so 
    that future synchronizations can just use "bzr colo-sync-from".  To change
    the saved synchronization location, use the "--remember" option.
    """

    takes_args = ['location?']

    takes_options = [option.Option('overwrite',
                                   help='Ignore differences between branches '
                                        'and overwrite unconditionally'),
                    'remember',
                    ]


    def run(self, location=None, overwrite=False, remember=False):
        workspace = ColocatedWorkspace()

        config = workspace.get_config()
        if location is None:
            # no specified location, must successfully get a saved location
            location = config.get_sync_from_location()
            if location is None:
                raise NoSavedSyncLocation
            remote_workspace = ColocatedWorkspace(location)
        else:
            # location specified
            remote_workspace = ColocatedWorkspace(location)
            if remember:
                config.set_sync_from_location(remote_workspace.base)
            else:
                if config.get_sync_from_location() is None:
                    # save first value as default
                    config.set_sync_from_location(remote_workspace.base)

        conflicts = 0
        current = workspace.current_branch_name()
        for name in remote_workspace.branch_names():
            remote_branch = remote_workspace.branch_with_name(name)
            try:
                local_branch = workspace.branch_with_name(name)
            except errors.NotBranchError:
                local_branch = None

            if local_branch is not None:
                trace.note('Updating branch %s from colo:%s:%s.' % (name,
                                                                    location,
                                                                    name))
                result = local_branch.pull(remote_branch, overwrite=overwrite)
                result.report(self.outf)

                if name == current:
                    # update the working tree
                    tree = workingtree.WorkingTree.open_containing(u'.')[0]
                    conflicts = tree.update()
            else:
                trace.note('Creating new branch %s from colo:%s:%s.' %
                           (name, location, name))
                branch_cmd = _wrap_command_object(builtins.cmd_branch(),
                                                  self.outf)
                branch_cmd.run(remote_branch.base,
                    to_location=workspace.url_for_name(name), no_tree=True)

        if conflicts > 0:
            return 1
        else:
            return 0


class cmd_colo_sync_to(commands.Command):

    """Synchronize this colocated workspace to another one.
    
    LOCATION should be a path or URL for a colocated workspace.  The branches
    in the current workspace are pushed to the specified remote workspace.
    If that path does not already contain matching branches, they will be
    created.

    The first time colo-sync-to is used, it records a default location so
    that future synchronizations can just use "bzr colo-sync-to".  To change
    the saved synchronization location, use the "--remember" option.
    """

    takes_args = ['location?']

    takes_options = [option.Option('overwrite',
                                   help='Ignore differences between branches '
                                        'and overwrite unconditionally.'),
                     'remember',
                    ]

    def run(self, location=None, overwrite=False, remember=False):
        workspace = ColocatedWorkspace()

        config = workspace.get_config()
        if location is None:
            # no specified location, must successfully get a saved location
            location = config.get_sync_to_location()
            if location is None:
                raise NoSavedSyncLocation
            remote_workspace = ColocatedWorkspace(location)
        else:
            # location specified
            remote_workspace = ColocatedWorkspace(location)
            if remember:
                config.set_sync_to_location(remote_workspace.base)
            else:
                if config.get_sync_to_location() is None:
                    # save first value as default
                    config.set_sync_to_location(remote_workspace.base)

        for name in workspace.branch_names():
            remote_branch_location = remote_workspace.url_for_name(name)
            trace.note('Updating colo:%s:%s from %s.' % (location,
                                                         name,
                                                         name))
            local_branch = workspace.branch_with_name(name)
            bzr_push_cmd = _wrap_command_object(builtins.cmd_push(), 
                                                self.outf)
            bzr_push_cmd.run(
                directory=urlutils.local_path_from_url(local_branch.base),
                location=remote_branch_location,
                overwrite=overwrite,
                create_prefix=True)




from unittest import TestLoader, TestSuite

import test_colo, test_sync, test_qcommands

def test_suite():
    suite = TestSuite()
    for t in [test_colo, test_sync, test_qcommands]:
        suite.addTests(TestLoader().loadTestsFromModule(t))
    return suite




# Copyright 2009 Neil Martinsen-Burrell

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 

from bzrlib.plugins.colo.colocated import (
    COLOCATED_LOCATION,
    ColocatedDirectory,
    ColocatedWorkspace,
    NoColocatedWorkspace,
    )
from bzrlib.plugins.colo.commands import (init_workspace,
                                          COLOCATED_STAGING,
                                         )
from bzrlib.plugins.colo.tests.script import ScriptTestCase

from bzrlib.lazy_import import lazy_import
lazy_import(globals(), '''
from bzrlib import (directory_service,
                    errors,
                    tests,
                    osutils,
                    urlutils,
                   )
''')


class TestColoLookup(tests.TestCaseWithTransport):

    def test_colo_lookup_fail(self):
        """Look up a colo: URL while not in a colocated workspace"""
        directory = ColocatedDirectory()
        self.assertRaises(NoColocatedWorkspace,
                          directory.look_up, '', 'colo:')

    def test_colo_lookup_nonexistent(self):
        """Look up a nonexistent branch under colo:"""
        directory = ColocatedDirectory()
        init_workspace(self.test_dir, outf=tests.StringIOWrapper())
        self.assertEqual(directory.look_up('test', 'colo:test'),
                         urlutils.local_path_to_url(osutils.pathjoin(
                             self.test_dir, COLOCATED_LOCATION, 'test')))

    def test_colo_in_directory_services(self):
        """Look up a branch via directory services."""
        self.assertRaises(NoColocatedWorkspace,
                          directory_service.directories.dereference, 
                          'colo:test')

    def test_colo_remote_error(self):
        self.assertRaises(NoColocatedWorkspace,
                          directory_service.directories.dereference,
                          'colo:test:remote')

    def test_colo_remote(self):
        self.run_bzr('colo-init test')
        self.assertEqual(directory_service.directories.dereference(
                         'colo:test:remote'),
                          urlutils.join(self.get_url(), 'test', COLOCATED_LOCATION, 'remote'))

    def test_error_code(self):
        self.run_bzr_error(['No colocated workspace'],
                           ['colo-branches'])

    def test_backslash_name(self):
        """A name with a backslash should be converted to a forward slash"""
        self.run_bzr('colo-init')
        self.assertEqual(directory_service.directories.dereference(
                         r'colo:test\backslash'),
                         urlutils.join(self.get_url(), COLOCATED_LOCATION,  'test/backslash'))



class TestColocatedWorkspace(tests.TestCaseInTempDir):

    def test_creation(self):
        init_workspace(u'.', outf=tests.StringIOWrapper())
        z = ColocatedWorkspace()
        self.assertStartsWith(z.base,
                              urlutils.local_path_to_url(self.test_dir))

    def test_creation_directory(self):
        init_workspace(u'branch', outf=tests.StringIOWrapper())
        z = ColocatedWorkspace('branch')
        self.assertEndsWith(z.base, 'branch/')

    def test_repo_location(self):
        init_workspace(u'.', outf=tests.StringIOWrapper())
        z = ColocatedWorkspace()
        self.assertEqual(z.repo_location, 
                         urlutils.local_path_to_url(
                             urlutils.join(self.test_dir, COLOCATED_LOCATION))
                        )

    def test_remote_no_tree_repo_location(self):
        init_workspace(u'.', create_tree=False, outf=tests.StringIOWrapper())
        z = ColocatedWorkspace(urlutils.normalize_url(self.test_dir))
        self.assertEqual(z.repo_location,
                         urlutils.normalize_url(
                             urlutils.join(self.test_dir, COLOCATED_LOCATION))
                        )

    def test_branch_names(self):
        init_workspace(u'.', outf=tests.StringIOWrapper())
        z = ColocatedWorkspace()
        self.assertEqual(list(z.branch_names()), ['trunk'])

    def test_current_branch(self):
        init_workspace(u'.', outf=tests.StringIOWrapper())
        z = ColocatedWorkspace()
        self.assertEndsWith(z.current_branch().base, 'trunk/')

    def test_current_branch_name(self):
        init_workspace(u'.', outf=tests.StringIOWrapper())
        z = ColocatedWorkspace()
        self.assertEqual(z.current_branch_name(), 'trunk')


class TestColoScripts(ScriptTestCase):

    def test_colo_blackbox(self):
        """colo: from the command line"""
        self.run_script("""
$ bzr colo-init
$ bzr init colo:test
$ cat .bzr/branches/test/.bzr/README
""")

    def test_colo_remote_blackbox(self):
        self.run_script("""
$ bzr colo-init b1
$ bzr colo-init b2
$ cd b1
$ bzr log colo:../b2:trunk
""")

    def test_colo_remote_url(self):
        script_text = '\n'.join(["$ bzr colo-init",
                                 "$ bzr log colo:%s:trunk" % self.get_url(),
                                    ])
        self.run_script(script_text)

    def test_colo_init(self):
        """Creating a colocated branch using the alias."""
        self.run_script("""
$ bzr colo-init branch
$ cat branch/.bzr/README
$ cat branch/.bzr/branches/.bzr/README
$ cat branch/.bzr/branches/trunk/.bzr/README
""")

    def test_branches(self):
        self.run_script("""
$ bzr colo-init
$ bzr colo-branch fix1
$ bzr colo-branch --no-switch fix2
""")
        z = ColocatedWorkspace()
        self.assertEqual(len(list(z.branches())), 3)

    def test_hierarchical_names(self):
        self.run_script("""
$ bzr colo-init
$ bzr colo-branch a/b
$ bzr colo-branches
* a/b
  trunk
""")
        z = ColocatedWorkspace()
        self.assertEqual(z.current_branch_name(), 'a/b')

    def test_backslash_names(self):
        self.run_script("""
$ bzr colo-init
$ bzr colo-branch trunk\\test
$ bzr colo-branches
  trunk
* trunk/test
$ bzr log colo:trunk\\test
""")

    def test_backslash_bialix(self):
        self.run_script("""
$ bzr colo-init
$ bzr ci -m1 --unchanged
$ bzr switch -b colo:trunk\\bar
$ bzr colo-branches
  trunk
* trunk/bar
$ bzr ci -m2 --unchanged
$ bzr switch colo:trunk
$ bzr merge colo:trunk\\bar
""")

    def test_colo_init_hidden(self):
        out, err = self.run_bzr(['colo-init'])
        self.assertNotContainsRe(out,
                                 COLOCATED_STAGING)

    def test_colo_init_format(self):
        """Creating a colocated branch using the alias."""
        self.run_script("""
$ bzr colo-init --format=1.9 branch
$ cat branch/.bzr/README
$ cat branch/.bzr/branches/.bzr/README
$ cat branch/.bzr/branches/trunk/.bzr/README
""")

    def test_colo_init_create_prefix(self):
        self.run_script("""
$ bzr colo-init --create-prefix branch/test/long/path
$ cat branch/test/long/path/.bzr/README
""")

    def test_colo_init_trunk_name(self):
        self.run_script("""
$ bzr colo-init --trunk-name=HEAD
$ bzr colo-branches
* HEAD
""")

    def test_colo_init_trunk_name_hierarchical(self):
        self.run_script("""
$ bzr colo-init --trunk-name=origin/trunk
$ bzr colo-branches
* origin/trunk
""")

    def test_colo_init_no_tree(self):
        self.run_script("""
$ bzr colo-init --no-tree
$ bzr colo-branches
  trunk
""")

    def test_colo_checkin(self):
        """Use a colocated branch to check in under trunk."""
        self.run_script("""
$ bzr colo-init
$ cat > file1
<text
$ cat < file1
text
$ bzr add
$ bzr ci -m 'first check-in'
2>Committing to: .../.bzr/branches/trunk/
2>added file1
2>Committed revision 1.
$ bzr log -r1
------------------------------------------------------------
revno: 1
committer: ...
branch nick: trunk
timestamp: ...
message:
  first check-in
""")

    def test_colo_switch_creates_branch(self):
        """Branch using the switch command in a colocated tree"""
        self.run_script("""
$ bzr colo-init
$ bzr switch -b newtip
$ cat .bzr/branches/newtip/.bzr/README
$ bzr nick
newtip
""")

    def test_colo_switch_creates_branch_use_directory_service(self):
        """Branch using the switch command in a colocated tree"""
        self.run_script("""
$ bzr colo-init
$ bzr switch -b colo:newtip/  # slash for workaround on switch bug
$ cat .bzr/branches/newtip/.bzr/README
$ bzr nick
newtip
""")

    def test_colo_merging(self):
        """Merge in colocated branches"""
        self.run_script("""
$ bzr colo-init
$ cat > file1
<text
$ bzr add
$ bzr ci -m 'adding first file'
$ bzr switch -b newtip
$ bzr nick
newtip
$ cat >> file1
<more text
$ cat < file1
text
more text
$ bzr ci -m 'change in newtip'
2>Committing to: .../.bzr/branches/newtip/
2>modified file1
2>Committed revision 2.
$ bzr switch trunk
$ bzr merge colo:newtip
$ bzr ci -m 'merge in changes from newtip'
2>Committing to: .../.bzr/branches/trunk/
2>modified file1
2>Committed revision 2.
$ bzr log --line -n0
2: ... [merge] merge in changes from newtip
  1.1.1: ... change in newtip
1: ... adding first file
""")

class TestColoBranch(ScriptTestCase):

    def test_colo_branch_then_switch(self):
        """Branch in a colocated tree."""
        self.run_script("""
$ bzr colo-init
$ bzr branch . colo:newtip
$ bzr switch newtip  # works because newtip is a sibling of trunk
$ bzr nick
newtip
""")

    def test_colo_branch(self):
        self.run_script("""
$ bzr colo-init
$ bzr colo-branch newtip
2>Branched 0 revision(s).
2>Tree is up to date at revision 0.
2>Switched to branch: .../.bzr/branches/newtip/
$ bzr nick
newtip
""")

    def test_colo_branch_no_switch(self):
        self.run_script("""
$ bzr colo-init
$ bzr colo-branch --no-switch newtip
2>Branched 0 revision(s).
$ bzr nick
trunk
""")

    def test_colo_branch_from(self):
        self.run_script("""
$ bzr colo-init
$ bzr ci -m 'commit in trunk' --unchanged
$ bzr colo-branch new
$ bzr ci -m 'commit in new branch' --unchanged
$ bzr colo-branch --from-branch=trunk new2
$ bzr log --line
1: ...
""")

    def test_colo_branch_revision(self):
        self.run_script("""
$ bzr colo-init
$ cat >file1
<content
$ bzr add
$ bzr ci -m 'first'
$ cat >>file1
<more content
$ bzr ci -m 'second'
$ bzr colo-branch -r1 new
$ cat file1
content
""")


class TestColoBranches(ScriptTestCase):

    def test_colo_branches(self):
        self.run_script("""
$ bzr colo-init
$ bzr colo-branch new
$ bzr colo-branch newer
$ bzr colo-branches
  new
* newer
  trunk
""")

    def test_colo_branches_location(self):
        self.run_script("""
$ bzr colo-init test
$ cd test
$ bzr colo-branch new
$ cd ..
$ bzr colo-branches test
  new
  trunk
""")


class TestColoFetch(ScriptTestCase):

    def test_fetch(self):
        self.make_branch('branch')
        self.run_script("""
$ bzr colo-fetch branch branch-colo
$ cd branch-colo
$ bzr colo-branches
* origin/trunk
""")

    def test_fetch_format(self):
        self.make_branch('branch')
        self.run_script("""
$ bzr colo-fetch --format=1.9 branch branch-colo
$ cd branch-colo
$ bzr colo-branches
* origin/trunk
""")

    def test_fetch_current_directory(self):
        self.make_branch('branch')
        self.run_script("""
$ bzr colo-fetch branch
2>bzr: ERROR: File exists: u'.../branch/.bzr'
""")

    def test_fetch_one_argument(self):
        self.run_script("""
$ bzr init --create-prefix other/branch
$ bzr colo-fetch other/branch
$ bzr colo-branches branch
  origin/trunk
""")

    def test_fetch_url(self):
        self.make_branch('branch')
        branch_url = urlutils.local_path_to_url('branch')
        self.run_script(
            ''.join(["$ bzr colo-fetch %s .\n" % branch_url,
                     "$ bzr colo-branches\n",
                     "* origin/trunk"]))

    def test_fetch_branch_name(self):
        self.make_branch('branch')
        self.run_script("""
$ bzr colo-fetch --trunk-name=master branch .
$ bzr colo-branches
* master
""")


class TestColoCheckout(ScriptTestCase):

    def test_checkout(self):
        self.run_script("""
$ bzr colo-init branch
$ cd branch
$ bzr colo-checkout ../branch2
$ cd ../branch2
$ cat .bzr/README
""")

    def test_checkout_example(self):
        self.run_script("""
$ bzr colo-init branch
$ cd branch
$ bzr colo-branch fix-eol-bug
$ bzr colo-branches
* fix-eol-bug
  trunk
$ bzr switch trunk
$ bzr colo-branch fix-unicode-normalization
$ bzr colo-checkout ../fix-unicode-normalization
$ cd ../fix-unicode-normalization
$ cat .bzr/README
""")

    def test_checkout_create(self):
        self.run_script("""
$ bzr colo-init branch
$ cd branch
$ bzr colo-checkout --create ../fix1
$ bzr colo-branches
  fix1
* trunk
$ bzr nick
trunk
$ cd ../fix1
$ cat .bzr/README
$ bzr nick
fix1
$ bzr colo-branches
* fix1
  trunk
""")

    def test_checkout_create_dash_b(self):
        self.run_script("""
$ bzr colo-init branch
$ cd branch
$ bzr colo-checkout -b ../fix1
$ bzr colo-branches
  fix1
* trunk
$ bzr nick
trunk
$ cd ../fix1
$ cat .bzr/README
$ bzr nick
fix1
$ bzr colo-branches
* fix1
  trunk
""")

    def test_checkout_in_directory(self):
        self.run_script("""
$ bzr colo-init
$ bzr colo-checkout tree
$ cd tree
$ cat .bzr/README
$ bzr nick
trunk
""")

    def test_checkout_from_branch(self):
        self.run_script("""
$ bzr colo-init branch
$ cd branch
$ bzr colo-branch fix-eol-bug
$ bzr colo-branches
* fix-eol-bug
  trunk
$ bzr colo-checkout --from-branch trunk ../fix-unicode-normalization
$ cd ../fix-unicode-normalization
$ bzr nick
trunk
$ bzr colo-branches
  fix-eol-bug
* trunk
""")


class TestColoPrune(ScriptTestCase):

    def test_prune(self):
        self.run_script("""
$ bzr colo-init
$ bzr colo-branch new
$ bzr colo-branches
* new
  trunk
$ bzr switch trunk
$ bzr colo-branches
  new
* trunk
$ bzr colo-prune new
$ bzr colo-branches
* trunk
""")

    def test_prune_multiple(self):
        self.run_script("""
$ bzr colo-init
$ bzr colo-branch --no-switch new
$ bzr colo-branch --no-switch newer
$ bzr colo-prune new newer
$ bzr colo-branches
* trunk
""")

    def test_prune_clean(self):
        self.run_script("""
$ bzr colo-init
$ bzr ci -m 'trunk revision' --unchanged
$ bzr colo-branch new
$ bzr ci -m 'a revision' --unchanged
$ bzr switch trunk
$ bzr colo-prune --clean new
$ bzr colo-branches
* trunk
""")
        repo = ColocatedWorkspace().repository()
        self.assertEqual(len(repo.all_revision_ids()), 1)

    def test_prune_current(self):
        self.run_script("""
$ bzr colo-init
$ bzr colo-prune trunk
2>Not removing current branch trunk.
""")

    def test_dont_prune_everything(self):
        self.run_script("""
$ bzr colo-init
$ bzr colo-branch one
$ bzr colo-prune .
2>Not removing . because it is not a branch.
$ cat .bzr/branches/.bzr/README
""")

    def test_dont_prune_subdirectories(self):
        self.run_script("""
$ bzr colo-init
$ bzr colo-branch origin/trunk
$ bzr colo-prune origin
2>Not removing origin because it is not a branch.
$ cat .bzr/branches/origin/trunk/.bzr/README
""")

    def test_prune_colo_prefix(self):
        self.run_script("""
$ bzr colo-init
$ bzr colo-branch new
$ bzr colo-prune colo:trunk
$ bzr colo-branches
* new
""")


class TestColoClean(ScriptTestCase):

    def test_clean(self):
        self.run_script("""
$ bzr colo-init
$ bzr ci -m 'trunk revision' --unchanged
$ bzr colo-branch new
$ bzr ci -m 'a revision' --unchanged
$ bzr switch trunk
$ bzr colo-prune new
$ bzr colo-clean
""")
        repo = ColocatedWorkspace().repository()
        self.assertEqual(len(repo.all_revision_ids()), 1)

    def test_clean_early_branch(self):
        self.run_script("""
$ bzr colo-init
$ bzr ci -m 'first revision' --unchanged
$ bzr ci -m 'second revision' --unchanged
$ bzr colo-branch -r 1 new
$ bzr ci -m 'new revision' --unchanged
$ bzr switch trunk
$ bzr colo-prune new
$ bzr colo-clean
""")
        self.assertEqual(
            len(ColocatedWorkspace().repository().all_revision_ids()), 2)

    def test_clean_merge(self):
        self.run_script("""
$ bzr colo-init
$ bzr ci -m 'first revision' --unchanged
$ bzr ci -m 'second revision' --unchanged
$ bzr colo-branch -r 1 new
$ bzr ci -m 'new revision' --unchanged
$ bzr switch trunk
$ bzr merge colo:new
$ bzr ci -m 'merge revision'
$ bzr colo-prune new
$ bzr colo-clean
""")
        # clean shouldn't remove anything
        self.assertEqual(
            len(ColocatedWorkspace().repository().all_revision_ids()), 4)

    def test_clean_very_old_format(self):
        self.run_script("""
$ bzr colo-init --knit
$ bzr colo-clean
2>bzr: ERROR: Can't clean the repository at .../.bzr/branches
""")

    def test_clean_old_format(self):
        self.run_script("""
$ bzr colo-init --pack-0.92
$ bzr colo-clean
""")


class TestColoIfy(ScriptTestCase):

    def test_colo_ify(self):
        self.run_script("""
$ bzr init test
$ cd test
$ bzr colo-ify
""")

    def test_colo_ify_change_name(self):
        self.run_script("""
$ bzr init test
$ cd test
$ bzr colo-ify --trunk-name=HEAD
$ bzr colo-branches
* HEAD
""")

    def test_colo_ify_others(self):
        self.run_script("""
$ bzr init test
$ bzr init test2
$ cd test
$ bzr colo-ify ../test2
$ bzr colo-branches
  test2
* trunk
""")

    def test_colo_ify_check_no_parent(self):
        self.run_script("""
$ bzr init test
$ cd test
$ bzr colo-ify
""")
        out, err = self.run_bzr(['info'])
        self.assertNotContainsRe(out, 'parent')

    def test_colo_ify_check_parent(self):
        self.run_script("""
$ bzr init test
$ bzr branch test test2
$ bzr branch test2 test3
$ cd test
$ bzr colo-ify ../test2 ../test3
$ bzr info colo:test3
...
  parent branch: .../test2
""")
        

class TestColoFixup(ScriptTestCase):

    def test_fixup_is_needed(self):
        self.run_script("""
$ bzr colo-init test
$ mv test test2
$ cd test2
""")
        self.run_bzr(['colo-branches'], retcode=3, 
                     error_regexes=['Not a branch:'])

    def test_fixup(self):
        self.run_script("""
$ bzr colo-init test
$ mv test test2
$ cd test2
$ bzr colo-fixup
$ bzr colo-branches
* trunk
""")

    def test_fixup_multiple(self):
        self.run_script("""
$ bzr colo-init test
$ cd test
$ bzr colo-branch feature
$ cd ..
$ mv test test2
$ cd test2
$ bzr colo-fixup
$ bzr colo-branches
* feature
  trunk
""")

    def test_fixup_hierarchical(self):
        self.run_script("""
$ bzr colo-init test
$ cd test
$ bzr colo-branch a/b
$ cd ..
$ mv test test2
$ cd test2
$ bzr colo-fixup
$ bzr colo-branches
* a/b
  trunk
""")

    def test_fixup_hierarchical_from_trunk(self):
        self.run_script("""
$ bzr colo-init test
$ cd test
$ bzr colo-branch a/b
$ bzr switch colo:trunk
$ cd ..
$ mv test test2
$ cd test2
$ bzr colo-fixup
$ bzr colo-branches
  a/b
* trunk
""")

    def test_fixup_from_subdir(self):
        self.run_script("""
$ bzr colo-init test
$ cd test
$ mkdir subdir
$ cd ..
$ mv test test2
$ cd test2/subdir
$ bzr colo-fixup
$ bzr colo-branches
* trunk
""")


class TestHierarchicalNames(ScriptTestCase):

    def test_branch(self):
        self.run_script("""
$ bzr colo-init
$ bzr colo-branch a/b
$ bzr colo-branch a/c
$ bzr colo-branch a/c/d
""")

    def test_branches(self):
        self.run_script("""
$ bzr colo-init
$ bzr colo-branch a/b
$ bzr colo-branches
* a/b
  trunk
""")

    def test_checkout(self):
        self.run_script("""
$ bzr colo-init test
$ cd test
$ bzr colo-branch a/b
$ bzr colo-checkout ../b-checkout
$ cd ../b-checkout
$ bzr switch colo:trunk
""")

    def test_prune(self):
        self.run_script("""
$ bzr colo-init
$ bzr colo-branch a/b
$ bzr switch colo:trunk
$ bzr colo-prune a/b
$ bzr colo-branches
* trunk
""")


class TestColoPull(ScriptTestCase):

    def test_pull(self):
        self.run_script("""
$ bzr colo-init workspace
$ cd workspace
$ bzr colo-branch origin/test
$ bzr colo-pull
2>Updating origin/test from .../trunk/
No revisions to pull.
""")

    def test_pull_outside_branch(self):
        self.run_script("""
$ bzr init branch
$ bzr colo-init workspace
$ cd workspace
$ bzr colo-branch origin/test
$ bzr pull --remember ../branch
$ bzr colo-pull
2>Updating origin/test from .../branch/
No revisions to pull.
""")

    def test_pull_revision(self):
        self.run_script("""
$ bzr init branch
$ bzr colo-init workspace
$ cd workspace
$ bzr colo-branch origin/test
$ bzr pull --remember ../branch
$ cd ../branch
$ echo "contents" > file
$ bzr add
$ bzr ci -m 'first'
$ cd ../workspace
$ bzr colo-pull
Now on revision 1.
""")

    def test_pull_diverged(self):
        self.run_script("""
$ bzr init branch
$ bzr colo-init workspace
$ cd workspace
$ bzr colo-branch origin/test
$ bzr pull --remember ../branch
$ cd ../branch
$ echo "contents" > file
$ bzr add
$ bzr ci -m 'first'
$ cd ../workspace
$ echo "other contents" > file2
$ bzr add
$ bzr ci -m 'a new revision'
""")
        out, err = self.run_bzr('colo-pull', retcode=3)
        self.assertEqual(out, '')
        self.assertContainsRe(err, 'bzr: ERROR: These branches have diverged.'
                              ' Use the missing command to see how.\n'
                              'Use the merge command to reconcile them.\n')

    def test_pull_multiple(self):
        self.run_script("""
$ bzr init branch
$ bzr colo-init workspace
$ cd workspace
$ bzr colo-branch origin/test
$ bzr pull --remember ../branch
$ bzr colo-branch origin/other
$ bzr pull --remember ../branch
$ cd ../branch
$ echo "contents" > file
$ bzr add
$ bzr ci -m 'first'
$ cd ../workspace
$ bzr colo-pull
2>Updating origin/other from .../branch/
2>All changes applied successfully.
2>Updating origin/test from .../branch/
Now on revision 1.
Now on revision 1.
""")

    def test_pull_current(self):
        self.run_script("""
$ bzr init branch
$ bzr colo-init workspace
$ cd workspace
$ bzr colo-mv trunk origin/trunk
$ bzr pull --remember ../branch
$ cd ../branch
$ echo content > file
$ bzr add
$ bzr ci -m 'a revision on branch'
$ cd ../workspace
$ bzr colo-pull
Now on revision 1.
$ cat file
content
""")


class TestColoMv(ScriptTestCase):

    def test_mv(self):
        self.run_script("""
$ bzr colo-init
$ bzr colo-branches
* trunk
$ bzr colo-mv trunk test
$ bzr colo-branches
* test
""")

    def test_mv_hierarchical(self):
        self.run_script("""
$ bzr colo-init
$ bzr colo-branches
* trunk
$ bzr colo-mv trunk origin/trunk
$ bzr colo-branches
* origin/trunk
""")

    def test_mv_hierarchical_existing(self):
        self.run_script("""
$ bzr colo-init
$ bzr colo-branch origin/test
$ bzr colo-mv trunk origin/trunk
""")

    def test_mv_not_current(self):
        self.run_script("""
$ bzr colo-init
$ bzr colo-branch new
$ bzr colo-mv trunk HEAD
$ bzr colo-branches
  HEAD
* new
""")


class TestNoTree(ScriptTestCase):

    def test_list_branches(self):
        self.run_script("""
$ bzr colo-init --no-tree
$ bzr colo-branches
  trunk
""")

    def test_branch_current(self):
        self.run_script("""
$ bzr colo-init --no-tree
$ bzr colo-branch new
2>bzr: ERROR: Cannot find a current branch in a no-tree colocated workspace.
""")

    def test_branch_from(self):
        self.run_script("""
$ bzr colo-init --no-tree
$ bzr colo-branch --from=trunk new
$ bzr colo-branches
  new
  trunk
""")

    def test_pull(self):
        self.run_script("""
$ bzr colo-init remote
$ bzr colo-init --no-tree --trunk-name=origin/trunk local
$ cd local
$ bzr branch colo:../remote:trunk colo:origin/remote
$ bzr colo-pull
2>Updating origin/remote from .../remote/.bzr/branches/trunk/
""")

    def test_checkout(self):
        self.run_script("""
$ bzr colo-init --no-tree branch
$ cd branch
$ bzr colo-checkout ../checkout
2>bzr: ERROR: Cannot find a current branch in a no-tree colocated workspace.
""")

    def test_prune(self):
        self.run_script("""
$ bzr colo-init --no-tree
$ bzr colo-branch --from=trunk new
$ bzr colo-prune new
""")

    def test_mv(self):
        self.run_script("""
$ bzr colo-init --no-tree
$ bzr colo-mv trunk HEAD
""")

    def test_clean(self):
        self.run_script("""
$ bzr colo-init --no-tree
$ bzr colo-clean
""")

    def test_fixup(self):
        self.run_script("""
$ bzr colo-init --no-tree first
$ mv first second
$ cd second
$ bzr colo-fixup
2>bzr: ERROR: Not a branch: ...
""")



# Copyright 2009 Neil Martinsen-Burrell
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 

"""Work with colocated branches using current technology.

In order to provide a faster and simpler working model, this plugin tries
to support a configuration similar to git and Mercurial's colocated branches,
where there is a single working tree that can be switched between multiple
branches that all co-exist in the same directory.  This working model is
entirely possible using Bazaar's existing technology, and this plugin aims to
make it as simple as possible to use that model.  

This plugin does not add any new formats or objects to Bazaar, it simply
provides convenience commands for working with a certain *convention* for
branch storage.  This plugin provides the following commands

    * colo-init
    * colo-branch
    * colo-branches
    * qbranches (if the qbzr plugin is installed)
    * colo-fetch
    * colo-pull
    * colo-checkout
    * colo-prune
    * qprune (if the qbzr plugin is installed)
    * colo-mv
    * colo-clean
    * colo-ify
    * colo-fixup
    * colo-sync-from
    * colo-sync-to

For more information on working with this plugin, see ``bzr help
colo-tutorial`` and the help for the individual commands.

Referring to Colocated Branches
-------------------------------

This plugin adds a directory service of "colo:" that can be used to refer to
branches inside of the ``.bzr/branches`` directory.  So, for example we could
``bzr switch colo:fix-eol-bug`` to do work in the colocated ``fix-eol-bug``
branch.  Note that in some cases, the "colo:" prefix is unnecessary because
Bazaar automatically looks for branches in the directory where the current
branch is located.  So the previous command would also work with ``bzr switch
fix-eol-bug`` because switch searches sibling directories automatically.  (If
hierarchical branch names such as ``bugs/1024`` are used, then simply doing
``bzr switch trunk`` may fail, while ``bzr switch colo:trunk`` will always
work.)

From within a colocated workspace, the specifier ``colo:branch_name`` refers
to the branch with that name in the current colocated workspace.  It is
possible to refer to colocated branches in *other* workspaces using the syntax
``colo:workspace_location:branch_name``.  The ``workspace_location`` in this
form can be either a path (e.g. ``colo:../other_project:trunk``) or a URL
(e.g. ``colo:bzr+ssh://hostname/path/to/workspace:trunk``).

"""


import sys, os

from bzrlib import directory_service
from bzrlib.commands import plugin_cmds
from bzrlib.help_topics import topic_registry

from bzrlib.plugins.colo.info import bzr_plugin_version as version_info

def test_suite():
    from unittest import TestSuite
    import bzrlib.plugins.colo.tests
    res = TestSuite()
    res.addTest(bzrlib.plugins.colo.tests.test_suite())
    return res

directory_service.directories.register_lazy('colo:',
    'bzrlib.plugins.colo.colocated', 'ColocatedDirectory',
    'Easy access to colocated branches')

topic_registry.register_lazy("colo-tutorial",
    "bzrlib.plugins.colo.tutorial", "_colo_tutorial",
    "Tutorial on colocated branches with the colo plugin")

lazy_commands = [('cmd_colo_init', []),
                 ('cmd_colo_branch', []),
                 ('cmd_colo_branches', []),
                 ('cmd_colo_fetch', []),
                 ('cmd_colo_pull', []),
                 ('cmd_colo_checkout', ['colo-co']),
                 ('cmd_colo_mv', ['colo-move', 'colo-rename']),
                 ('cmd_colo_prune', ['colo-delete']),
                 ('cmd_colo_clean', []),
                 ('cmd_colo_ify', []),
                 ('cmd_colo_fixup', []),
                 ('cmd_qprune', ['qdelete']),
                 ('cmd_qbranches', []),
                 ('cmd_colo_sync_from', []),
                 ('cmd_colo_sync_to', []),
                ]
for cmd, aliases in lazy_commands:
    plugin_cmds.register_lazy(cmd, aliases, 'bzrlib.plugins.colo.commands')



# Copyright 2010 Neil Martinsen-Burrell

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 

import os
from bzrlib.tests import TestCaseWithTransport
from bzrlib import osutils

class TestQBzrCommands(TestCaseWithTransport):

    def test_disabled(self):
        # Incantation from MBP 20100413
        self.run_bzr_subprocess('qbranches', retcode=3,
                                env_changes=dict(BZR_DISABLE_PLUGINS='qbzr'),
                                allow_plugins=True)

    def test_enabled(self):
        try:
            from bzrlib.plugins import qbzr
        except ImportError:
            return
        self.run_bzr_error(['No colocated workspace'],
                           ['qbranches'])



_colo_tutorial = \
"""
=========================================
Colocated Branches with the `colo` plugin
=========================================

Basic Concepts
==============

Colocated Workspaces
--------------------

I'll refer to the amalgamation of a working tree and some number of related
branches within that same directory as a colocated workspace.  A colocated
workspace for developing on Bazaar itself would be a single directory named
``bazaar``.  It would consist of a single working tree in the ``bazaar``
directory, and one or more branches stored in a shared repository located at
``bazaar/.bzr/branches``.  This location is inside Bazaar's control directory
``bazaar/.bzr``, but the branches in that directory can still be referred to
just like any Bazaar branches.  The default branch is called ``trunk``.

Referring to Colocated Branches
-------------------------------

This plugin adds a directory service of "colo:" that can be used to
refer to branches inside of the ``.bzr/branches`` directory.  So, for example
we could ``bzr switch colo:fix-eol-bug`` to do work in the colocated
``fix-eol-bug`` branch.  Note that in some cases, the "colo:" prefix is
unnecessary because Bazaar automatically looks for branches in the directory
where the current branch is located.  So the previous command would also work
with ``bzr switch fix-eol-bug`` because switch searches sibling directories 
automatically.

From within a colocated workspace, the specifier ``colo:branch_name`` refers
to the specified branch in the current colocated workspace.  It is possible
to refer to colocated branches in arbitrary workspaces using the syntax 
``colo:workspace_location:branch_name``.  The ``workspace_location`` in this
form can be either a path (e.g. ``colo:../other_project:trunk``) or a URL
(e.g. ``colo:bzr+ssh://hostname/path/to/workspace:trunk``).

Branch names may use the chracter "/" to create hierarchical storage for
branches.  For example, one might store all branches related to bug fixes with
names that start with ``bugs/``.  This corresponds to the actual branches being
stored in a subdirectory of ``.bzr/branches``.

Inspecting Colocated Branches
-----------------------------

Since Bazaar doesn't have a built-in command for listing the branches in a
given repository, this plugin provides the ``colo-branches`` command to list
all of the branches in the ``.bzr/branches`` repository.  This is very similar
to the ``branches`` command provided by the bzrtools__ plugin.  If the current
directory is a checkout of one of the colocated branches, the current
branch is marked with a "*".  If the qbzr__ plugin is installed, then this
plugin also defines a graphical ``qbranches`` command to show all of the 
colocated branches, with the current one marked in bold.

__ http://launchpad.net/bzrtools
__ http://launchpad.net/qbzr

Also, since the ``.bzr/branches`` directory is an ordinary Bazaar shared
repository, tools such as ``bzr qlog`` from the qbzr plugin which are able
to show the log information for all of the branches found in a repository
can be easily used.  Just do ``bzr qlog colo:`` to see all of the
history for all of the branches in this colocated workspace.  Individual
branches can be logged simply by using the "colo:" specifier: ``bzr log
colo:trunk``.


Creating a Colocated Workspace
------------------------------

This plugin provides a ``colo-init`` command which creates a new colocated
workspace with a single working tree that is a checkout of a single branch
called "trunk".  This can then be the starting point for either a new project,
or for pulling in the sources of an existing project.  Within this directory,
"colo:" branch names can be used as the targets of commands such as ``bzr
branch`` and ``bzr pull`` in the process of building up a colocated workspace
from existing branches.

The colo-init command has an option ``--no-tree`` that creates colocated
workspaces with no working tree.  In this case, there will just be a
collection of branches in a shared repository at ``.bzr/branches``.  This can
be useful for workspaces created remotely where it isn't possible to make a
working tree and for any time where the working files aren't needed.

This plugin also offers a ``colo-ify`` command to convert an existing branch
into a colocated workspace.  That command can also take additional branch
arguments to add to the workspace.  For example, consider a situation where we
have branches in separate directories ``master/`` and ``fix-eol-bug/``.  Then,
the following commands would convert the master branch into a colocated
workspace with two branches inside::

    $ cd master
    $ bzr colo-ify ../fix-eol-bug 
    $ bzr colo-branches
      fix-eol-bug
    * trunk

Making Another Colocated Branch
-------------------------------

In the colocated workflow, a common operation is to create a new branch and
then switch the working tree to that new branch.  Using standard Bazaar
commands in a colocated workspace, this is a two step process::

    $ bzr branch . colo:new-branch
    $ bzr switch colo:new-branch

Since this is such a common process in the colocated workflow, this plugin
provides a command ``colo-branch`` that simplifies this process, condensing
the two commands above to::

    $ bzr colo-branch new-branch

Note that by default colo-branch branches from the current branch, so one
should be careful that the current branch is the correct one, or use the
``--from-branch`` option to specify the name of a source branch.  (Often ``bzr
branch --from-branch=trunk new_branch`` is the correct thing to do.)

Making a New Working Tree from a Workspace
------------------------------------------

Sometimes it is nice to be able to make a second working tree for the same
colocated workspace, for example to work on a second bug fix in parallel.
This plugin provides the ``colo-checkout`` command which creates a second
checkout of the current branch in the specified directory.  Often, the
specified directory will be a sibling of the current workspace.  For example::

    $ bzr colo-branches
      fix-eol-bug
    * trunk
    $ bzr colo-branch fix-unicode-normalization # creates a colocated branch
    $ bzr colo-checkout ../fix-unicode-normalization
    $ cd ../fix-unicode-normalization
    $ bzr colo-branches
      fix-eol-bug
    * fix-unicode-normalization
      trunk

Note that while the new checkout isn't a full colocated workspace (it doesn't
contain branches inside ``.bzr/branches``), it is still possible to use the
"colo:" directory specifier to refer to other branches in the original
colocated workspace.  For example, to see the effect of the unicode
normalization fixes, one could do::

    $ bzr diff --old=colo:trunk

Since creating a new colocated branch for work in a new checkout is a common
operation, the ``colo-checkout`` command takes a ``--create`` option
to indicate that a new colocated branch should be created and then the new
checkout should be switched to that new branch.  The name of the new branch
comes from the basename of the directory that is being created.
For example, we could shorten the above example as::

    $ bzr colo-checkout --create ../fix-unicode-normalization

As an additional simplification, the ``colo-checkout`` command also takes a
``--from-branch`` option to simplify the process of making additional
checkouts from branches other than the current branch.  Following the above
example, if we wanted to create the fix-unicode-normalization branch off of
the fix-eol-bug branch, we could do::

    $ bzr colo-checkout --create --from-branch=fix-eol-bug ../fix-unicode-normalization

Fetching an Existing Project
----------------------------

This plugin provides a ``colo-fetch`` command for getting an existing branch
into a new colocated workspace.  This can typically be used as the first step
in starting to do work on an existing project.  For example, to begin work on
developing Bazaar, one could do::
 
  $ bzr colo-fetch lp:bzr ~/src/bzr
  $ cd ~/src/bzr
  $ bzr colo-branches
  * origin/trunk

The ``colo-fetch`` command creates a single colocated branch named
"origin/trunk" by default.  If a different name is wanted, then the
``--trunk-name`` option can be specified.

To add additional branches to the colocated workspace, use the ordinary
``branch`` command and specify a branch name using the "colo:" specifier.  To
add Bazaar's stable 2.0 branch to the above colocated workspace would be::
  
  $ cd ~/src/bzr
  $ bzr branch lp:bzr/2.0 colo:origin/2.0
  $ bzr colo-branches
    origin/2.0
  * origin/trunk

This will reuse the existing revisions from the trunk branch and should
provide for a very quick download.  Branches whose name start with "origin/"
have a special meaning to the ``colo-pull`` command.

Updating Multiple Branches
--------------------------

This plugin provides a ``colo-pull`` command for updating all of the remote
branches that are mirrored in this workspace.  Since it often isn't desired to
pull every single branch from its parent, this plugin uses a simple convention
for indicating which branches should be updated by ``colo-pull``: all branches
whose name begins with "origin/" will be pulled from their parent locations.
By default, the ``colo-fetch`` command uses "origin/trunk" for the name of its
initial branch, indicating that it should be updated by ``colo-pull``.
Merging the "origin/" branches into the other local branches that have been
made from them must be done manually.

Using the colocated workspace set up in the previous section for Bazaar 
development using two branches, the colo-pull command would show the
following::

    $ bzr colo-pull
    Updating origin/2.0 from bzr+ssh://bazaar.launchpad.net/~bzr-pqm/bzr/2.0/
    Now on revision ...
    Updating origin/trunk from bzr+ssh://bazaar.launchpad.net/~bzr-pqm/bzr/bzr.dev/
    Now on revision ...

Renaming a Colocated Branch
---------------------------

Since colocated branches exist inside Bazaar's hidden control directory, it
can be inconvenient to move those branches using ordinary filesystem tools.
This plugin provides a ``colo-mv`` command to rename colocated branches. 
This command simply moves the branch on disk.  As a result, relationships 
between the moved branch and other checkouts or between the moved branch and 
other branches may be severed.  These relationships must be manually restored
after moving a branch, so use this command with caution.

An important use of this command is to designate a branch as remote by giving
it a name that starts with "origin/".  If the "trunk" branch should now be a
mirror of a remote branch, say on Launchpad at lp:project, then the following
would make that connection::

    $ bzr colo-branches
    ...
    * trunk
    $ bzr pull --remember lp:project
    $ bzr colo-mv trunk origin/trunk
    $ bzr colo-branches
    ...
    * origin/trunk
    $ bzr colo-pull
    Updating origin/trunk from bzr+ssh://bazaar.launchpad.net/~project/trunk
    No revisions to pull.

Synchronizing Colocated Workspaces
----------------------------------

For purposes of backup or moving work from one computer to another, one may
want to synchronize all of the branches in one colocated workspace with all
of the branches in another colocated workspace.  This plugin provides the
``colo-sync-to`` and ``colo-sync-from`` commands for this purpose.  These
commands take the location of another colocated workspace as an argument and
perform the synchronization in the specified direction using Bazaar's native
push and pull facilities.

For example, consider the case of synchronizing work on the local machine with
a corresponding workspace on a USB key at ``/mnt/usb_key``.  At the end of a
workday, one would want to update the workspace on the USB key to match that
of the local workspace::

    $ bzr colo-init /mnt/usb_key/bzr   # only necessary once
    $ cd ~/src/bzr
    $ bzr colo-sync-to /mnt/usb_key/bzr
    Updating colo:/mnt/usb_key/bzr:trunk from trunk
    ...

Then, if one did some work using their USB key overnight, then in the morning
they could sync from there to their local workspace::

    $ cd ~/src/bzr
    $ bzr colo-sync-from /mnt/usb_key/bzr
    Updating trunk from colo:/mnt/usb_key/bzr:trunk

Note that the colo-sync-from and colo-sync-to commands are symmetric in the
sense that both of them will update all of the branches that exist in the
*source* workspace.  This means that newly created branches in the local
workspace will also be created in the remote workspace upon ``colo-sync-to``
and newly created remote branches will be created in the local workspace when
using ``colo-sync-from``.  Neither of these commands will delete any branches,
but branches can be deleted in any workspace with ``colo-prune``.

For remote workspaces accessed over a network, Bazaar may refuse to create
working trees, but these workspaces can still be used for synchronization by
creating them with ``colo-init --no-tree``.  For example::

    $ bzr colo-init --no-tree bzr+ssh://hostname/path/to/backup
    $ bzr colo-sync-to bzr+ssh://hostname/path/to/backup

The ``colo-sync-to`` and ``colo-sync-from`` commands each save a default
location the first time they are used.  To synchronize somewhere other than
the default location, simply specify that location in the command, as 
was done on the first invocation.  To change the default location,
give the ``--remember`` option when specifying a new location.  Even if
you want to synchronize from and to the same workspace, you have to 
specify that location once to both ``colo-sync-from`` and ``colo-sync-to``:
the commands don't make use of each other's defaults.

Removing a Colocated Branch
---------------------------

Because colocated branches are stored in a relatively hidden directory, this
plugin provides a ``colo-prune`` command for removing a colocated branch.
It is equivalent to removing the branch's directory on the filesystem.  If 
the qbzr plugin is installed, then there is a graphical equivalent command
called ``qprune``.

To remove the revisions that were only present in the pruned branches, the
``--clean`` option can be specified.  This may be a long-running operation
if the history of the project is large.  In the case where one or more
``colo-prune`` operations has already happened, there is a ``colo-clean``
command that will remove all of the unused revisions.

Moving Colocated Workspaces on Disk
-----------------------------------

Because Bazaar uses absolute path names to record the connection between a
working tree and a branch, when colocated workspaces are moved around on the
filesystem, the connection between the working tree and the current branch is
broken.  To fix this, run ``bzr colo-fixup`` from within the working tree.  If
this doesn't work for some reason, then one can always run the command ``bzr
switch --force .bzr/branches/<current branch>``, provided one knows what the
current branch should be.


Example Usages
==============

Develop an existing project in a colocated workspace
----------------------------------------------------

We want to develop a GUI feature on the Bazaar project, using colocated
branches.

::

    $ bzr colo-fetch lp:bzr bzr
    $ cd bzr
    $ bzr colo-branches
    * origin/trunk
    $ bzr colo-branch feature-gui
    # make changes
    $ bzr ci -m 'add the new feature'
    ...
    $ bzr missing colo:origin/trunk
    You have 1 extra revision(s):
    ------------------------------------------------------------
    revno:...
    ...
    # submit the change, in this case using a merge directive
    $ bzr send colo:origin/trunk

Starting a new project using a colocated workspace
--------------------------------------------------

We want to start a new project and be ready to use colocated branches going
forward with the development.

::

    $ bzr colo-init newproj
    Shared repository (format: 2a)
    Location:
      shared repository: newproj/.bzr/branches
    Created a repository branch (format: 2a)
    Using shared repository: /tmp/newproj/.bzr/branches/
    $ cd newproj
    $ echo "I'm new!" > README.txt
    $ bzr add
    adding README.txt
    $ bzr ci -m 'first revision'
    Committing to: /tmp/newproj/.bzr/branches/trunk/
    added README.txt
    Committed revision 1.
    $ bzr log colo:trunk
    ------------------------------------------------------------
    revno: 1
    committer: ...
    branch nick: trunk
    timestamp: ...
    message:
      first revision!
    
Converting an existing project to a colocated workspace
-------------------------------------------------------

We want to take an existing project with a trunk and numerous bug fix branches
and convert it into a single colocated workspace.  Assume that we have the main
branch in the ``master/`` directory and bug fix branches in directories named
``fix-001/`` and ``fix-002/`.  We can convert the ``master/`` directory to a
colocated workspace using ``colo-ify``::

    $ cd master
    $ bzr colo-ify
    $ bzr colo-branches
    * trunk

If we want to use a different name for the trunk branch, we can use the
``--trunk-name`` option.

If we would now like the bug fix branches also as colocated branches, we can
simply branch them in using standard Bazaar commands::

    $ bzr branch ../fix-001 colo:fix-001
    $ bzr branch ../fix-002 colo:fix-002
    $ bzr colo-branches
      fix-001
      fix-002
    * trunk

We could also have specified the bug fix branches as arguments to the
``colo-ify`` command to specify bringing in those additional branches all at
once, making the preceding commands equivalent to::

    $ bzr colo-ify ../fix-001 ../fix-002

The new branches are created with names derived from their locations ("fix-001"
and "fix-002" in this case, as above).

Using two checkouts with a colocated workspace
----------------------------------------------

One of the advantages of a colocated workspace is having a single directory
which will always contain the active code for a project, for example when
using an IDE or when there are generated files that are expensive to
re-create.  This can be a disadvantage when we want to have two separate
working trees, for example to compare two sets of files or to maintain a
working tree that is up to date with a particular branch.  We can do this
using a colocated workspace and the ``colo-checkout`` command.

For example, when developing a project such as Bazaar, it can be important to
have a pristine copy of the trunk from which to run Bazaar while developing so
that the development work doesn't prevent working with the version control
system.

::

    $ bzr colo-fetch lp:bzr bzr
    $ cd bzr
    $ bzr colo-checkout ../bzr.dev
    $ bzr colo-branch fix-doc-bugs
    $ bzr colo-branches
    * fix-doc-bugs
      origin/trunk
    $ ../bzr.dev/bzr --version  # this runs the trunk version of bzr
    $ ./bzr --version  # this runs the version in the fix-doc-bugs branch

Another use for the ``colo-checkout`` command is to create a second working
tree to do some work without interrupting the work that is happening in the
colocated workspace (although a judicious use of ``bzr shelve`` can sometimes
make this unnecessary).  Here is an example that uses the ``--create`` option
to create a new branch and make a new checkout of it::

    $ bzr colo-branches
    * fix-doc-bugs
      origin/trunk
    $ bzr colo-checkout --create ../urgent-fix
    $ bzr colo-branches
    * fix-doc-bugs
      origin/trunk
      urgent-fix
    $ cd ../urgent-fix
    $ bzr colo-branches
      fix-doc-bugs
      origin/trunk
    * urgent-fix

Note that the original working tree still points at the same branch while only
the new checkout was switched to the newly created branch.


Limitations
===========

1. Colocated workspaces that are moved on disk need to have their checkout
   reconnected to its branch, due to the fact that Bazaar checkouts refer to
   absolute paths.  Do this using::
   
     $ bzr switch --force .bzr/branches/trunk

   from the root directory of the checkout, or try the ``colo-fixup`` command
   which should work in most cases.
"""



# Copyright 2009 Neil Martinsen-Burrell
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 

import os

from bzrlib.lazy_import import lazy_import
lazy_import(globals(), '''
from bzrlib import (branch as _mod_branch,
                    bzrdir,
                    config,
                    errors,
                    ignores,  
                    osutils,
                    repository,
                    trace,
                    transport,
                    urlutils, 
                   )
''')

COLOCATED_LOCATION = osutils.pathjoin(u'.bzr', u'branches')

class NoColocatedWorkspace(errors.BzrCommandError):

    _fmt = "No colocated workspace in %(directory)s"


class NoCurrentBranch(errors.BzrCommandError):

    _fmt = "Cannot find a current branch in a no-tree colocated workspace."


class ColocatedWorkspace(object):

    """An abstract representation of a colocated workspace."""

    def __init__(self, location=u'.'):
        """A colocated workspace that contains the given path or URL."""
        try:
            a_bzrdir, extra_path = bzrdir.BzrDir.open_containing(location)
        except errors.NotBranchError:
            raise NoColocatedWorkspace(directory=location)
        self.has_tree = a_bzrdir.has_workingtree()

        if not self.has_tree:
            self.base = urlutils.normalize_url(location)
        else:
            branch = a_bzrdir.open_branch()
            try:
                self.base = branch.base[:branch.base.index(COLOCATED_LOCATION)]
            except ValueError:
                raise NoColocatedWorkspace(directory=location)

        self.repo_location = urlutils.join(self.base, COLOCATED_LOCATION)

    def url_for_name(self, name):
        """Return a URL to the named branch."""
        # normalize backslashes to forward slashes here
        if '\\' in name:
            trace.mutter('Replacing backslash in %s' % name)
            name = name.replace('\\', '/')

        return urlutils.join(self.repo_location, name)

    def branches(self):
        """A generator of the branches in the colocated workspace."""
        t = self.repo_transport()
        return (b for b in bzrdir.BzrDir.find_branches(t) 
                if b.base.startswith(t.base))

    def _name_from_branch(self, branch):
        """A name for the given branch object."""
        return branch.base[len(self.repo_location):].strip('/')

    def repo_transport(self):
        """A transport pointing to the shared repository."""
        return transport.get_transport(self.repo_location)

    def repository(self):
        """The shared repository object."""
        return repository.Repository.open(self.repo_location)

    def branch_names(self):
        """A generator of the names of the branches."""
        for b in self.branches():
            yield self._name_from_branch(b)

    def branch_with_name(self, name):
        return _mod_branch.Branch.open(self.url_for_name(name))

    def current_branch(self, directory=u'.'):
        """The current branch object.

        If this workspace doesn't have a working tree and thus doesn't have a
        current branch, raise an exception.
        """
        if not self.has_tree:
            raise NoCurrentBranch()
        return _mod_branch.Branch.open_containing(directory)[0]

    def current_branch_name(self, directory=u'.'):
        """The name of the current branch.

        If this workspace doesn't have a working tree and thus doesn't have
        a current branch, return None.
        """
        try:
            return self._name_from_branch(self.current_branch(directory))
        except NoCurrentBranch:
            return None

    def get_config(self):
        return ColocatedConfig(self)


class ColocatedDirectory(object):

    """Directory lookup for branches that are colocated with the tree.
    
    If the name portion contains a ":", then interpret the name as
    workspace_location:branch_name.  If not, interpret name as a branch in the
    current colocated workspace. 
    """

    def look_up(self, name, url):
        url_piece, sep, branch_name = name.rpartition(':') 
        if not url_piece:
            url_piece = u'.'
        return ColocatedWorkspace(url_piece).url_for_name(branch_name)


class ColocatedConfig(config.BzrDirConfig):

    """Store configuration information for a colocated workspace."""

    def __init__(self, workspace):
        super(ColocatedConfig, self).__init__(bzrdir.BzrDir.open(workspace.base))

    def _get_option(self, option):
        if self._config is None:
            return None
        value = self._config.get_option(option)
        if value == '':
            value = None
        return value

    def get_sync_from_location(self):
        """Get the location to synchronize from"""
        return self._get_option('sync_from_location')

    def get_sync_to_location(self):
        """Get the location to synchronize to"""
        return self._get_option('sync_to_location')

    def set_sync_from_location(self, location):
        """Set the location to synchronize from"""
        if location is None:
            location = ''
        self._config.set_option(location, 'sync_from_location')

    def set_sync_to_location(self, location):
        """Set the location to synchronize to"""
        if location is None:
            location = ''
        self._config.set_option(location, 'sync_to_location')




# Copyright 2010 Neil Martinsen-Burrell

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 


import bzrlib
from bzrlib.lazy_import import lazy_import
lazy_import(globals(),'''
            from bzrlib import (errors,
                                tests,
                               )
            ''')


class DummyScriptTestCase(tests.TestCase):

    def setUp(self):
        raise tests.TestSkipped('Using bzr version < 2.1, '
                                'skipping script tests')


if bzrlib.version_info < (2, 1):
    ScriptTestCase = DummyScriptTestCase
else:
    from bzrlib.tests import script
    ScriptTestCase = script.TestCaseWithTransportAndScript



# Copyright 2010 Neil Martinsen-Burrell
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 

import os
from bzrlib.lazy_import import lazy_import
lazy_import(globals(), '''
from bzrlib import (commands,
                    errors,
                    option,
                    trace,
                   )
''')
from colocated import ColocatedWorkspace
from PyQt4 import QtCore, QtGui

from bzrlib.plugins.qbzr.lib.commands import QBzrCommand
from bzrlib.plugins.qbzr.lib.util import (QBzrWindow,
                                          BTN_CLOSE,
                                          runs_in_loading_queue,
                                         )
from bzrlib.plugins.qbzr.lib.uifactory import ui_current_widget
from bzrlib.plugins.qbzr.lib.trace import reports_exception
from bzrlib.plugins.qbzr.lib.subprocess import run_subprocess_command


class ColocatedBranchesList(QtGui.QListWidget):

    def __init__(self, workspace):
        super(ColocatedBranchesList, self).__init__()
        self.workspace = workspace
        self.setSelectionMode(QtGui.QAbstractItemView.ExtendedSelection)
        self.branch_icon = QtGui.QIcon(
            os.path.join(os.path.dirname(__file__), 'folder.png'))

    def refresh_branches(self):
        self.clear()
        current_name = self.workspace.current_branch_name()
        for name in reversed(sorted(list(self.workspace.branch_names()))):
            item = QtGui.QListWidgetItem(name)
            item.setIcon(self.branch_icon)
            item.setToolTip(self.workspace.branch_with_name(name).base)
            if name == current_name:
                bold = QtGui.QFont(); bold.setWeight(QtGui.QFont.Bold)
                item.setFont(bold)
            self.insertItem(-1, item)


class ColocatedBranchesWindow(QBzrWindow):

    def __init__(self, title=None, parent=None, ui_mode=True):
        super(ColocatedBranchesWindow, self).__init__(title=title,
                                                      parent=parent,
                                                      ui_mode=ui_mode)
        self.workspace = ColocatedWorkspace()
        self.branches_list = ColocatedBranchesList(self.workspace)

    def show(self):
        QBzrWindow.show(self)
        QtCore.QTimer.singleShot(1, self.initial_load)

    @runs_in_loading_queue
    @ui_current_widget
    @reports_exception()
    def initial_load(self):
        self.branches_list.refresh_branches()


class cmd_qprune(QBzrCommand):

    """Remove colocated branches."""

    takes_options = [option.Option('directory',
                     short_name='d',
                     type=unicode,
                     help='location of colocated workspace'),
                    ]

    def _qbzr_run(self, directory=None):
        if directory is not None:
            os.chdir(directory)
        self.main_window = PruneWindow()
        self.main_window.show()
        self._application.exec_()


class PruneWindow(ColocatedBranchesWindow):
    
    def __init__(self, title=None,
                       parent=None, ui_mode=True):
        super(PruneWindow, self).__init__(title=title, 
                                          parent=parent,
                                          ui_mode=ui_mode)
        if title is None:
            self.set_title(("Delete Colocated Branches", os.getcwdu()))  
        
        self.restoreSize("prune", (340, 220))

        self.layout = QtGui.QVBoxLayout(self.centralwidget)

        hbox = QtGui.QHBoxLayout()
        self.layout.addLayout(hbox)
        hbox.addWidget(self.branches_list)

        actions_box = QtGui.QVBoxLayout()
        hbox.addLayout(actions_box)
        prune_button = QtGui.QPushButton('Delete')
        actions_box.addWidget(prune_button, 0, QtCore.Qt.AlignTop)
        refresh_button = QtGui.QPushButton('Refresh')
        actions_box.addWidget(refresh_button, 1, QtCore.Qt.AlignTop)
        
        self.layout.addWidget(self.create_button_box(BTN_CLOSE))

        QtCore.QObject.connect(refresh_button, 
                               QtCore.SIGNAL("clicked(bool)"),
                               self.branches_list.refresh_branches)
        QtCore.QObject.connect(prune_button,
                               QtCore.SIGNAL("clicked(bool)"),
                               self.do_prune)
        
    def do_prune(self):
        for item in self.branches_list.selectedItems():
            branch_name = item.text()
            if branch_name == self.workspace.current_branch_name():
                self._warn_no_delete_current()
                continue
            run_subprocess_command('colo-prune %s' % branch_name)
        self.branches_list.refresh_branches()

    def _warn_no_delete_current(self):
        msg_box = QtGui.QMessageBox(
            QtGui.QMessageBox.Warning,
            'Warning',
            'Not deleting current branch: %s' %
            self.workspace.current_branch_name(),
            QtGui.QMessageBox.Ok,
            self)
        msg_box.exec_()


class cmd_qbranches(QBzrCommand):

    """List colocated branches.
    
    The current branch is marked in bold.
    """

    takes_options = [option.Option('directory', short_name='d',
                     type='unicode',
                     help='location of colocated workspace'),
                    ]

    def _qbzr_run(self, directory=None):
        if directory is not None:
            os.chdir(directory)

        self.main_window = BranchesWindow()
        self.main_window.show()
        self._application.exec_()


class BranchesWindow(ColocatedBranchesWindow):

    def __init__(self, title=None,
                       parent=None, ui_mode=True):
        super(BranchesWindow, self).__init__(title=title,
                                             parent=parent,
                                             ui_mode=ui_mode)
        if title is None:
            self.set_title(('Colocated Branches', os.getcwdu()))

        self.restoreSize("branches", (340, 220))
        
        self.layout = QtGui.QVBoxLayout(self.centralwidget)
        self.layout.addWidget(self.branches_list)
        buttonbox = self.create_button_box(BTN_CLOSE)
        refresh_button = buttonbox.addButton('Refresh',
                                             QtGui.QDialogButtonBox.ActionRole)
        self.layout.addWidget(buttonbox)

        QtCore.QObject.connect(refresh_button,
                               QtCore.SIGNAL('clicked(bool)'),
                               self.branches_list.refresh_branches
                              )



#!/usr/bin/env python
from distutils.core import setup

from info import (bzr_plugin_version,
                  bzr_plugin_name,
                  bzr_minimum_version,
                  bzr_commands,
                  __version__,
                 )

if __name__ == '__main__':
    setup(name="bzr-colo",
          version=__version__,
          description="Work with colocated Bazaar branches "
                      "using current technology.",
          author="Neil Martinsen-Burrell",
          author_email="nmb@wartburg.edu",
          url="https://launchpad.net/bzr-colo`",
          packages=['bzrlib.plugins.colo',
                    'bzrlib.plugins.colo.tests',
                    'bzrlib.plugins.colo.explorer',
                    ],
          package_dir={'bzrlib.plugins.colo': '.'},
          classifiers=['Development Status :: 4 - Beta',
                       'Intended Audience :: Developers',
                       'License :: OSI Approved :: GNU General Public License (GPL)',
                       'Programming Language :: Python :: 2',
                       'Topic :: Software Development :: Version Control',
                      ]
          
         )



"""Store version info separately for use in __init__.py and setup.py"""
bzr_plugin_name = 'colo'
bzr_plugin_version = (0, 2, 0, 'final', 0)
bzr_minimum_version = (2, 1, 0)

if bzr_plugin_version[3] == 'final':
    __version__ = '%d.%d.%d' % bzr_plugin_version[:3]
else:
    __version__ = '%d.%d.%d%s%d' % bzr_plugin_version

bzr_commands = ['colo-init',
                'colo-branch',
                'colo-branches',
                'colo-fetch',
                'colo-pull',
                'colo-checkout',
                'colo-co',
                'colo-mv',
                'colo-move',
                'colo-rename',
                'colo-prune',
                'colo-delete',
                'colo-clean',
                'colo-ify',
                'colo-fixup',
                'qprune',
                'qdelete',
                'qbranches',
                'colo-sync-from',
                'colo-sync-to',
               ]
